# Copyright 2020 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from typing import Literal
from typing import Callable
from typing import Optional
from contextvars import ContextVar
import threading
import queue
import logging

from embrace import exceptions
from embrace import poolvalidators

logger = logging.getLogger(__name__)

ConnType = Any


class ContextManagerWrappedConnection:
    def __init__(self, pool: "ConnectionPoolBase", transaction=False, contextual=True):
        self.conn = None
        self.pool = pool
        self.transaction = transaction
        self.contextual = contextual

    def __enter__(self) -> ConnType:
        self.conn = self.pool.getconn(self.contextual)
        return self.conn

    def __exit__(self, exc_type, exc_value, tb):
        if self.transaction:
            try:
                if exc_type:
                    self.conn.rollback()
                else:
                    self.conn.commit()
            finally:
                self.pool.release(self.conn)
        else:
            self.pool.release(self.conn)


class ConnectionPoolBase:
    """
    Connection pool base class
    """

    #: A callable returning a db-api connection object
    connection_factory: Callable[[], ConnType]

    #: The class instantiated in response to the ``connect`` method
    connection_context_manager = ContextManagerWrappedConnection

    #: How many simultaneous connections to allow. If zero the number will be
    #: unlimited
    limit: int

    #: Return the same connection in the same context (thread)
    contextual: bool = False
    connection_context: ContextVar[ConnType | None]

    #: Callable to release a connection. Must return True if the connection
    #: may be reused, else False.
    before_release: Optional[Callable[[ConnType], bool]]

    #: Callable called every time a connection is acquired to validate
    #: that it is still alive
    validate: Optional[Callable[[ConnType], bool]]

    def __init__(
        self,
        connection_factory,
        limit=0,
        validator: Callable[[ConnType], bool] | Literal["auto"] | None = "auto",
        contextual=False,
    ):
        if isinstance(validator, poolvalidators.ConnectionValidator):
            self.validate = validator.validate
            self.before_release = validator.before_release
        elif validator == "auto":
            self.validate = self.auto_validate  # type: ignore
            self.before_release = None
        else:
            self.validate = None
            self.before_release = None
        self.connection_factory = connection_factory  # type: ignore # noqa
        self.limit = limit
        self.contextual = contextual
        self.max_validation_retries = self.limit + 3
        self.connection_context = ContextVar("connection_context", default=None)

    def _getconn(self) -> ConnType:
        raise NotImplementedError()

    def getconn(self, contextual=True) -> ConnType:
        """
        Get a connection from the pool. The connection must be released by
        calling ``release``.
        """
        raise NotImplementedError

    def release(self, conn: ConnType):
        """
        Release the connection back to the pool
        """
        raise NotImplementedError

    def connect(self, transaction=False, contextual=True):
        """
        Return a context manager that manages acquiring and releasing a
        connection.

        :param transaction:
            If true, ``commit()`` will be called on the connection object at
            the end of the block, or ``rollback()`` in case of exception.
        """
        return self.connection_context_manager(self, transaction, contextual)

    def transaction(self, contextual=True):
        """
        A convenience method equivalent to calling
        :meth:`~embrace.pool.ConnectionPoolBase(transaction=True)`
        """
        return self.connection_context_manager(self, True, contextual)

    def set_validator(self, v):
        self.validate = v.validate
        self.before_release = v.before_release

    def auto_validate(self, conn):
        validator = poolvalidators.ConnectionValidator()

        for cls in poolvalidators.validators:
            if isinstance(conn, cls):
                validator = poolvalidators.validators[cls]()
                break
        self.set_validator(validator)
        return validator.validate(conn)


class ConnectionPool(ConnectionPoolBase):
    """
    A connection pool implementation using a queue to provide thread-safety.
    """

    # Maintain the pool in a queue for thread/process safety
    queue_class = queue.Queue
    lock_class = threading.Lock

    #: How long to wait for a connection to become available
    timeout = 5

    _pool: queue.Queue

    def __init__(
        self,
        connection_factory,
        limit=0,
        validator: Callable[[ConnType], bool] | Literal["auto"] | None = "auto",
        timeout=timeout,
        contextual=False,
    ):
        super(ConnectionPool, self).__init__(
            connection_factory, limit, validator, contextual=contextual
        )

        self._pool = self.queue_class()
        self.lock = self.lock_class()
        self.connections_created = 0
        self.reached_limit = False
        self.timeout = timeout

    def getconn(self, contextual=True) -> ConnType:
        contextual = contextual and self.contextual
        if contextual and (conncount := self.connection_context.get()):
            conn, refcount = conncount
            self.connection_context.set((conn, refcount + 1))
            return conn

        try:
            conn = self._pool.get(block=self.reached_limit, timeout=self.timeout)
        except queue.Empty:
            if self.limit:
                with self.lock:
                    if self.reached_limit:
                        raise exceptions.ConnectionLimitError()
                    else:
                        conn = self._connect()
            else:
                conn = self._connect()

        if self.validate:
            retry = 0
            while True:
                try:
                    if self.validate(conn):
                        break
                except Exception:
                    logger.exception(f"Could not validate {conn=}")

                logger.info(f"Invalid connection {conn}")
                self.release(conn)

                if retry >= self.max_validation_retries:
                    raise Exception(
                        f"Could not validate a connection after "
                        f"{self.max_validation_retries} attempts"
                    )

                conn = self._getconn()
                retry += 1

        if contextual:
            conncount = self.connection_context.get()
            assert conncount is None
            self.connection_context.set((conn, 1))
        return conn

    def _getconn(self):
        """
        Return a connection from the pool.
        """

    def _connect(self):
        self.connections_created += 1
        self.reached_limit = bool(self.limit and self.connections_created >= self.limit)
        return self.connection_factory()  # type: ignore

    def release(self, conn):

        full_release = True
        if self.contextual:
            conncount = self.connection_context.get()
            if conncount:
                context_conn, refcount = conncount
                if conn is context_conn:
                    if refcount > 1:
                        self.connection_context.set((conn, refcount - 1))
                        full_release = False
                    else:
                        self.connection_context.set(None)
                        full_release = True

        if full_release:
            try:
                reuse = self.before_release(conn) if self.before_release else True
            except Exception:
                reuse = False
                logger.exception(f"Could not release {conn=}")

            if reuse:
                self._pool.put(conn)
            else:
                conn.close()
                if self.limit:
                    self.lock.acquire()
                    self.connections_created -= 1
                    self.reached_limit = self.connections_created >= self.limit
                    self.lock.release()


class SQLAlchemyPool(ConnectionPoolBase):
    def __init__(self, connection_factory, limit=0, validator=None):
        if limit:
            raise ValueError(
                "Cannot set limit here, "
                "configure this in SQLAlchemy connection pooling instead"
            )
        if validator:
            raise ValueError(
                "Cannot set validator here, "
                "configure this in SQLAlchemy connection pooling instead"
            )

        def _connection_factory(sessiongetter=connection_factory):
            return sessiongetter().connection().connection.connection

        super(SQLAlchemyPool, self).__init__(_connection_factory, limit, validator)

    def getconn(self, contextual=True) -> ConnType:
        return self.connection_factory()  # type: ignore

    def release(self, conn):
        pass



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Sequence
from itertools import chain
from itertools import islice
from typing import Callable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union
from typing import TYPE_CHECKING
import dataclasses
import sys
import logging
import typing as t

import sqlparse

from .parsing import compile_bind_parameters
from .parsing import split_statements
from .parsing import compile_includes
from .parsing import get_placeholder_info
from .parsing import Metadata
from . import parsing
from .mapobject import mapobject
from .mapobject import make_rowspec
from .mapobject import joinedload
from .mapobject import RowMapper
from . import exceptions
from .types import Connection
from .types import Cursor


if TYPE_CHECKING:
    from embrace.module import Module

known_styles: dict[type, str] = {}

_marker = object()
T = t.TypeVar("T")
U = t.TypeVar("U")
V = t.TypeVar("V")

logger = logging.getLogger(__name__)


def get_param_style(conn: Connection, known_styles=known_styles) -> str:
    conncls = conn.__class__
    try:
        return known_styles[conncls]
    except KeyError:
        modname = conncls.__module__
        while modname:
            try:
                style = sys.modules[modname].paramstyle  # type: ignore
                known_styles[conncls] = style
                return style
            except AttributeError:
                if "." in modname:
                    modname = modname.rsplit(".", 1)[0]
                else:
                    break
    raise TypeError(f"Can't find paramstyle for connection {conn!r}")


class ResultSet(t.Generic[T, U]):
    __slots__ = ["cursor", "row_mapper", "maprows"]

    def __init__(self, cursor, row_mapper: RowMapper):
        self.cursor = cursor
        self.row_mapper = row_mapper
        if row_mapper:
            self.maprows = self.row_mapper.make_row_mapper()
        else:
            self.maprows = None

    def __iter__(self):
        return self.many()

    def one(self) -> T:
        row = self.cursor.fetchone()
        if row is None:
            raise exceptions.NoResultFound()
        if self.cursor.fetchone() is not None:
            raise exceptions.MultipleResultsFound()
        if self.maprows:
            return next(self.maprows([row]))  # type: ignore
        return row

    def first(self) -> T:
        row = self.cursor.fetchone()
        if row and self.maprows:
            return next(self.maprows([row]))  # type: ignore
        return row

    def many(
        self, method: str = "auto", limit_rows: int = 0, limit_mapped: int = 0
    ) -> t.Iterable[T]:
        """
        Return an iterator over the ResultSet.

        :param method: The cursor method used to fetch results.
                       Must be one of 'auto', 'fetchone', 'fetchmany', or 'fetchall'
                       (default: 'auto'). If ``auto`` is selected, ``fetchall``
                       will be used unless a limit is given, in which case
                       ``fetchmany`` will be used.
        :param limit_rows: Maximum number of rows to fetch.
                           This is based on the number of rows returned. If you
                           are using joins to compose these into objects, this
                           may not correspond with the number of objects yielded.
        :param limit_mapped: Maximum number of mapped objects to generate.
        """
        if method == "auto":
            # Shortcut the default case
            if not limit_mapped and not limit_rows:
                if self.maprows:
                    return self.maprows(self.cursor.fetchall())
                else:
                    return iter(self.cursor.fetchall())
            method = "fetchmany"

        row_getter = getattr(self, f"many_rows_{method}")
        if self.maprows:
            if limit_mapped:

                def limited_rowmapper():
                    count = 0
                    rowcount = 0

                    if method == "fetchmany":

                        def _row_getter():
                            while True:
                                if count == 0:
                                    limit = self.cursor.arraysize
                                else:
                                    remaining = limit_mapped - count
                                    limit = int(((remaining + 1) * rowcount) / count)

                                ix = 0
                                for row in row_getter(limit=limit):
                                    yield row
                                    ix += 1
                                if ix == 0:
                                    return

                        rows = _row_getter()
                    else:
                        rows = row_getter()

                    for ix, mapped in enumerate(self.maprows(rows), 1):
                        yield mapped
                        if ix >= limit_mapped:
                            break

                return limited_rowmapper()
            else:
                return self.maprows(row_getter(limit=limit_rows))
        else:
            return row_getter(limit=(limit_rows or limit_mapped))

    def many_rows_fetchone(self, limit: int = 0) -> t.Iterator[T]:
        if limit:
            return islice(iter(self.cursor.fetchone, None), limit)
        else:
            return iter(self.cursor.fetchone, None)

    def many_rows_fetchmany(self, limit: int = 0) -> t.Iterator[T]:
        if limit:
            arraysize = self.cursor.arraysize
            fullbatches, remainder = divmod(limit, arraysize)
            for ix in range(fullbatches):
                yield from self.cursor.fetchmany(arraysize)
            if remainder:
                yield from self.cursor.fetchmany(remainder)
        else:
            for batch in iter(self.cursor.fetchmany, []):
                yield from batch

    def many_rows_fetchall(self, limit: int = 0) -> t.Iterator[T]:
        if limit:
            return iter(self.cursor.fetchall()[:limit])
        else:
            return iter(self.cursor.fetchall())

    def one_or_none(self) -> t.Optional[T]:
        row = self.cursor.fetchone()
        if self.cursor.fetchone() is not None:
            raise exceptions.MultipleResultsFound()
        if row and self.maprows:
            return next(self.maprows([row]))
        return row

    @t.overload
    def scalar(self, default: V) -> t.Union[U, V]:
        ...

    @t.overload
    def scalar(self) -> U:
        ...

    def scalar(self, default=_marker):
        value = next(self.column(), default)
        if value is _marker:
            raise exceptions.NoResultFound()
        return value

    def column(self) -> t.Iterator[T]:
        if self.row_mapper:
            column_slice = self.row_mapper.get_split_points()[0]
            row_mapper = dataclasses.replace(
                self.row_mapper, row_spec=self.row_mapper.row_spec[:1]
            )
            rows = (row[column_slice] for row in iter(self.cursor.fetchone, None))
            return row_mapper.make_row_mapper()(rows)
        first = self.cursor.fetchone()
        if first:
            if isinstance(first, Mapping):
                key = next(iter(first))
            elif isinstance(first, Sequence):
                key = 0
            else:
                raise TypeError(
                    f"Can't find first column for row of type {type(first)}"
                )
            return (
                row[key] for row in chain([first], iter(self.cursor.fetchone, None))
            )
        return iter([])

    def affected(self) -> int:
        return self.cursor.rowcount

    @property
    def rowcount(self) -> int:
        return self.cursor.rowcount


class Query(t.Generic[T, U]):
    name: Optional[str]
    source: Optional[str]
    lineno: Optional[int]
    includes: List[str] = []
    includes_resolved = False
    metadata: Metadata = Metadata()
    statements: list[str]
    row_mapper: Optional[RowMapper] = None

    def __init__(
        self,
        sql: Optional[str] = None,
        source: Optional[str] = None,
        lineno: Optional[int] = None,
        name: Optional[str] = None,
        metadata: Optional[Metadata] = None,
    ):
        self.name = name
        self.result_map = None
        self.src = sql
        self.statements = []
        if sql:
            self._configure_from_string(sql)

        self.metadata = metadata or self.metadata
        self.source = source
        self.lineno = lineno
        self._conn = None

    def __repr__(self):
        return f"<{self.__class__.__name__} from {self.source}:{self.lineno}>"

    @classmethod
    def from_statements(
        cls, metadata: Metadata, statements: Sequence[sqlparse.sql.Statement], **kwargs
    ) -> "Query":
        q = cls(**kwargs)
        q._configure_from_statements(metadata, statements)
        return q

    def _configure_from_string(self, s: str):
        metadata, statements = next(iter(split_statements(s)))
        self._configure_from_statements(metadata, statements)

    def _configure_from_statements(
        self, metadata: Metadata, statements: Sequence[sqlparse.sql.Statement]
    ):
        try:
            returning = parsing.parse_returning_metadata(
                imports=metadata.imports,
                returns=metadata.returns,
                joins=metadata.joins,
                key_columns=metadata.key_columns,
                split_on=metadata.split_on,
            )
        except Exception:
            logger.exception(f"Error parsing {metadata=}")
            raise
        self.statements = [str(s) for s in statements]
        self.metadata = metadata
        if returning:
            self._configure_returning(
                row_spec=returning.row_spec,
                joins=returning.joins,
                key_columns=returning.key_columns,
                split_on=returning.split_on,
            )

    def resolve_includes(self, module: "Module") -> bool:
        new_statements = []
        for stmt in self.statements:
            include_names, replace_includes = compile_includes(stmt)
            try:
                queries = [module.queries[n] for n in include_names]
            except KeyError:
                self.includes_resolved = False
                return False
            for q in queries:
                q.resolve_includes(module)

            new_statements.append(replace_includes(queries))
            self.includes = include_names
        self.statements = new_statements
        self.includes_resolved = True
        return True

    def get_placeholder_info(self) -> list[tuple[str, str]]:
        """
        Return a list of placeholder names in the query in source order
        """
        names = {}
        for sql in self.statements:
            for name, type_hint in get_placeholder_info(sql):
                names.setdefault(name, type_hint)
        return list(names.items())

    def copy(self) -> "Query":
        """
        Return a copy of the query
        """
        cls = self.__class__
        new = cls.__new__(cls)
        new.__dict__ = self.__dict__.copy()
        return new

    def bind(self, conn) -> "Query":
        """
        Return a copy of the query bound to a database connection
        """
        bound = self.copy()
        bound._conn = conn
        return bound

    def returning(
        self,
        row_spec: Union["mapobject", Callable, Sequence[Union["mapobject", Callable]]],
        joins: Union[joinedload, Sequence[joinedload]] = [],
        positional=False,
        key_columns: Optional[list[tuple[str, ...]]] = None,
        split_on=[],
    ) -> "Query":
        """
        Return a copy of the query configured to map and join objects

        :param row_spec:
            One of:
                - a single callable or :class:`mapobject` instance
                - a sequence of callables and :class:`mapobject` instances

            In the case that a single callable or mapobject is provided, the
            values for each row will be passed to the callable as keyword
            arguments, and each row will return a single python object.

            If a sequence of items is provided, the row will be split up into
            multiple objects at the columns indicated by the values of
            ``split_on``, or configured in :attr:`mapobject.split_on`.

        :param joins:
            When multiple python objects are generated per-row, ``joins``
            tells embrace how to construct relations between objects. See
            :func:`joinone` and :func:`joinmany`.

        :param positional:
            If True, pass column values to the items in ``row_spec`` as
            positional parameters rather than keyword args.

        :key_columns:
            The list of primary key columns for each object mapped by ``row_spec``.
            Rows with the same values in these columns will be mapped to the
            same python object.

            If key_columns is not provided, rows will be mapped to the same
            python object if all values match.

            It is an error to use this parameter if :class:`mapobject` is
            used anywhere in ``row_spec``.

        :split_on:
            List of the column names that mark where new objects start
            splitting a row into multiple python objects. If not provided,
            the string ``'id'`` will be used.

            It is an error to use this parameter if :class:`mapobject` is
            used anywhere in ``row_spec``.
        """
        q = self.copy()
        q._configure_returning(
            row_spec=row_spec,
            joins=joins,
            positional=positional,
            key_columns=key_columns,
            split_on=split_on,
        )
        return q

    def _configure_returning(
        self,
        row_spec: Union["mapobject", Callable, Sequence[Union["mapobject", Callable]]],
        joins: Union[joinedload, Sequence[joinedload]] = [],
        positional=False,
        key_columns: Optional[list[tuple[str, ...]]] = None,
        split_on=[],
    ):
        if isinstance(joins, joinedload):
            joins = [joins]

        row_spec = make_rowspec(row_spec, split_on or [], key_columns or [], positional)
        self.row_mapper = RowMapper(row_spec, joins=joins)

    def resultset(
        self,
        conn: Optional[Connection] = None,
        params: Optional[Mapping] = None,
        *,
        default=_marker,
        debug: bool = False,
        _result: Optional[str] = None,
        _get_param_style=get_param_style,
        _resultset=ResultSet,
        _compile_bind_parameters=compile_bind_parameters,
        **kw,
    ) -> ResultSet[T, U]:
        if conn is None:
            conn = self._conn
            if conn is None:
                raise TypeError("Query must be called with a connection argument")

        paramstyle = _get_param_style(conn)
        cursor: Cursor = conn.cursor()  # type: ignore
        if params:
            if kw:
                kw.update(params)
                params = kw
        else:
            params = kw

        prepared = [
            _compile_bind_parameters(paramstyle, s, params) for s in self.statements
        ]
        for sqltext, bind_params in prepared:
            if debug:
                import textwrap

                print(
                    f"Executing \n{textwrap.indent(sqltext, '    ')} "
                    f"with {bind_params!r}",
                    file=sys.stderr,
                )
            try:
                if bind_params:
                    cursor.execute(sqltext, bind_params)
                else:
                    cursor.execute(sqltext)
            except BaseException:
                _handle_exception(conn, self.source, 1)

        if self.row_mapper and cursor.description:
            self.row_mapper.set_description(cursor.description)
            row_mapper = self.row_mapper
        else:
            row_mapper = None
        return _resultset(cursor, row_mapper)

    def one(self, conn=None, *, debug=False, **kwargs) -> T:
        return self.resultset(conn, debug=debug, **kwargs).one()

    def one_or_none(self, conn=None, *, debug=False, **kwargs) -> t.Optional[T]:
        return self.resultset(conn, debug=debug, **kwargs).one_or_none()

    @t.overload
    def scalar(self, conn=None, *, default: V, debug=False, **kwargs) -> t.Union[U, V]:
        ...

    @t.overload
    def scalar(self, conn=None, *, debug=False, **kwargs) -> U:
        ...

    def scalar(self, conn=None, *, default=_marker, debug=False, **kwargs):
        return self.resultset(conn, debug=debug, **kwargs).scalar(default=default)

    def first(self, conn=None, *, debug=False, **kwargs) -> t.Optional[T]:
        return self.resultset(conn, debug=debug, **kwargs).first()

    def many(self, conn=None, *, debug=False, **kwargs) -> t.Iterable[T]:
        return self.resultset(conn, debug=debug, **kwargs).many()

    def column(self, conn=None, *, debug=False, **kwargs) -> Iterator[U]:
        return self(conn, debug=debug, **kwargs).column()  # type: ignore

    def cursor(self, conn=None, *, debug=False, **kwargs) -> Cursor:
        return self.resultset(conn, debug=debug, **kwargs).cursor

    def affected(self, conn=None, *, debug=False, **kwargs) -> int:
        return self.resultset(conn, debug=debug, **kwargs).rowcount

    __call__ = execute = resultset
    rowcount = affected


def _handle_exception(conn, filename, lineno):
    """
    We have an exception of unknown type, probably raised
    from the dbapi module
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    if exc_type and exc_value:
        classes = [exc_type]
        while classes:
            cls = classes.pop()
            clsname = cls.__name__

            if clsname in exceptions.pep_249_exception_names:
                newexc = exceptions.pep_249_exception_names[clsname]()
                newexc.args = getattr(exc_value, "args", tuple())
                newexc = newexc.with_traceback(exc_tb)
                if filename:
                    newlines = "\n" * (lineno - 1)
                    exec(
                        compile(
                            f"{newlines}raise newexc from exc_value", filename, "exec"
                        ),
                        {
                            "__name__": filename,
                            "__file__": filename,
                            "newexc": newexc,
                            "exc_value": exc_value,
                        },
                        {},
                    )
                else:
                    raise newexc from exc_value
            classes.extend(getattr(cls, "__bases__", []))

        raise exc_value.with_traceback(exc_tb) from exc_value


def joinone(target, attr, source):
    """
    Populate `<target>.<attr>` with
    the item identified by `source`.
    """
    return joinedload(target, attr, source, "1")


def joinmany(target, attr, source):
    """
    Populate `<target>.<attr>` with the list of items identified by `source`.
    """
    return joinedload(target, attr, source, "*")


one_to_many = joinmany
one_to_one = joinone



from collections.abc import Sequence
from collections import namedtuple
from datetime import date
from datetime import datetime
from decimal import Decimal
from functools import partial
from itertools import chain
from itertools import repeat
from itertools import pairwise
from itertools import zip_longest
from pickle import dumps
from pickle import HIGHEST_PROTOCOL
import typing as t
import dataclasses

from .types import CursorDescription

T = t.TypeVar("T")


class NullObjectType:
    pass


NullObject = NullObjectType()
joinedload = namedtuple("joinedload", "target attr source cardinality")
column_indexes = namedtuple(
    "column_indexes", "target_idx target_attr source_idx cardinality join_as_dict"
)

#: Common builtin / stdlib types that map to a single column
single_column_types = {bytes, bool, int, float, str, Decimal, datetime, date}


@dataclasses.dataclass(frozen=True, eq=True)
class mapobject:
    mapped: t.Callable
    key: dataclasses.InitVar[t.Union[str, Sequence[str]]] = tuple()
    split: str = "id"
    positional: bool = False
    column_count: t.Optional[int] = None
    join_as_dict: bool = False
    key_columns: Sequence[str] = tuple()
    label: t.Any = None

    @staticmethod
    def passthrough_mapped(x: T) -> T:
        return x

    def __post_init__(self, key):
        if not self.key_columns:
            if isinstance(key, str):
                object.__setattr__(self, "key_columns", (key,))
            else:
                object.__setattr__(self, "key_columns", (tuple(key)))
        if self.label is None:
            object.__setattr__(self, "label", self.mapped)
        if self.mapped in single_column_types:
            object.__setattr__(self, "positional", True)
            object.__setattr__(self, "positional", True)
            object.__setattr__(self, "column_count", 1)
        elif self.mapped is dict:
            object.__setattr__(self, "join_as_dict", True)

    @classmethod
    def dict(cls, mapped=dict, *args, **kwargs):
        kwargs["join_as_dict"] = True
        return cls(mapped, *args, **kwargs)

    @classmethod
    def passthrough(cls, mapped="ignore", *args, **kwargs):
        kwargs["column_count"] = 1
        kwargs["positional"] = True
        return cls(cls.passthrough_mapped, **kwargs)

    @classmethod
    def namedtuple(cls, *args, **kwargs):
        return cls(mapped=dynamicnamedtuple(), **kwargs)

    @classmethod
    def dataclass(
        cls,
        fields: Sequence[
            t.Union[
                str,
                tuple[str, t.Union[str, type]],
                tuple[str, t.Union[str, type], dataclasses.Field],
            ]
        ] = [],
        **kwargs: t.Any,
    ):
        """
        Return a mapper that stores values in a dataclass.

        By default field names are be inferred from the returned column names,
        and have a type of ``typing.Any``.

        :param fields:
            Additional fields to be appended after fields loaded from the
            database. This should be in the same format as expected by the
            stdlib :func:`dataclasses.make_dataclass` function.

            If a field has the same name as one of the returned
            columns, it will be used instead of autogenerating a field.

            If a field tuple does not contain a :class:`dataclasses.Field` item,
            the value ``dataclasses.field(default=None)`` will be used.
            Additional fields are not populated during data loading so
            must have a default or default_factory specified.

        :param kwargs:
            Keyword arguments corresponding to :class:`mapobject` fields will be
            passed through to the constructor (eg ``split``, ``label``, etc).
            Other arguments will be taken as additional field names, and
            specified as either a simple name-type mapping::

                mapobject.dataclass(foo=list[str]|None)

            or as a tuple of ``(type, Field)``::

                mapobject.dataclass(
                    foo=(list[str], dataclasses.field(default_factory=list))
                )
        """
        mapobject_fields = {f.name for f in dataclasses.fields(cls)}
        field_dict = {item[0]: item for item in fields}
        field_dict.update(
            (name, (name,) + (item if isinstance(item, tuple) else (item,)))
            for name, item in kwargs.items()
            if name not in mapobject_fields
        )
        mapobject_kwargs = {k: v for k, v in kwargs.items() if k in mapobject_fields}

        return cls(mapped=dynamicdataclass(field_dict), **mapobject_kwargs)


class dynamicnamedtuple:
    nt = None

    def __call__(self, **kwargs):
        nt = self.nt
        if nt is None:
            nt = self.nt = namedtuple(  # type: ignore
                f"mapobject_namedtuple_{id(self)}", tuple(kwargs)
            )
        return self.nt(**kwargs)  # type: ignore


class dynamicdataclass:
    datacls = None

    def __init__(self, field_dict):
        self.field_dict = field_dict

    def __call__(self, **kwargs):
        datacls = self.datacls
        if datacls is None:
            dbfields = [self.field_dict.pop(k, k) for k in kwargs]
            additionalfields = []
            for item in self.field_dict.values():
                if isinstance(item, str):
                    item = (item,)
                if len(item) < 3:
                    item = item + (dataclasses.field(default=None),)
                additionalfields.append(item)
            datacls = self.datacls = dataclasses.make_dataclass(
                f"mapobject_dataclass_{id(self)}",
                dbfields + additionalfields,
            )
        return datacls(**kwargs)


def make_rowspec(
    row_spec: t.Union[
        "mapobject", t.Callable, Sequence[t.Union["mapobject", t.Callable]]
    ],
    split_on: Sequence[str],
    key_columns: Sequence[tuple[str, ...]],
    positional: bool,
) -> list["mapobject"]:
    """
    Standardize a row_spec into a standardized sequence of mapobjects.
    :param row_spec:
        The row_spec as provided to Query.returning.
        This could be a single mapobject (eg ``mapobject.dict()``),
        a callable (eg ``dict``), or a sequence mixing the two styles.
    :param split_on:
        A sequence of column names at which to split mapped objects
    :param key_columns:
        A sequence of column name tuples uniquely identifying a
        mapped object.
    """
    if not isinstance(row_spec, Sequence):
        row_spec = (row_spec,)

    if split_on and any(isinstance(i, mapobject) for i in row_spec):
        raise TypeError("Cannot combine mapobject with split_on")

    result = []
    for ix, item in enumerate(row_spec):
        if not isinstance(item, mapobject):
            item = mapobject(
                mapped=item,
                positional=positional,
                split=(split_on[ix - 1] if 0 < ix < len(split_on) - 1 else "id"),
                key_columns=(key_columns[ix] if ix < len(key_columns) else []),
            )

        result.append(item)
    return result


@dataclasses.dataclass
class RowMapper:
    """
    Model the properties required to map a query's output (cursor description +
    row tuples) to the configured return type
    """

    row_spec: list[mapobject] = dataclasses.field(default_factory=list)
    imports: list[str] = dataclasses.field(default_factory=list)
    joins: t.Sequence[joinedload] = dataclasses.field(default_factory=list)
    key_columns: list[tuple[str, ...]] = dataclasses.field(default_factory=list)
    split_on: list[str] = dataclasses.field(default_factory=list)

    split_points: list[slice] = dataclasses.field(default_factory=list)
    is_multi: bool = False
    column_names: t.Optional[list[str]] = None
    mapped_column_names: t.Optional[list[tuple[str, ...]]] = None
    _row_mapper: t.Optional[t.Callable[[t.Iterable[tuple]], t.Iterator[t.Any]]] = None

    def __post_init__(self):
        self.is_multi = len(self.row_spec) > 1

    def set_description(self, description: CursorDescription):
        self.column_names = [d[0] for d in description]
        if self.is_multi:
            sp = self.get_split_points()
            self.split_points = sp
            self.mapped_column_names = [tuple(self.column_names[s]) for s in sp]
        else:
            self.mapped_column_names = [tuple(self.column_names)]

    def make_row_mapper(self) -> t.Callable[[t.Iterable[tuple]], t.Iterator[t.Any]]:
        if self._row_mapper:
            return self._row_mapper
        maker: t.Optional[t.Callable] = None
        grouper: t.Optional[t.Callable] = None
        if self.column_names is None:
            raise AssertionError("set_description must be called first")

        if self.joins:
            if not self.is_multi:
                raise TypeError(
                    "joins may only be set when there are multiple return types"
                )
            grouper = partial(
                group_by_and_join, self.mapped_column_names, self.row_spec, self.joins
            )

        else:
            maker = make_object_maker(self.mapped_column_names, self.row_spec)

        self._row_mapper = partial(self._maprows, grouper, maker)
        return self._row_mapper

    def row_spec_after_joins(self) -> list[mapobject]:
        """
        Return the row_spec after all joins have been applied, consolidating
        the output tuple
        """
        return [
            item
            for item in self.row_spec
            if not any(item.label == j.source for j in self.joins)
        ]

    def get_split_points(self) -> list[slice]:
        pos = 0
        result = []
        if self.column_names is None:
            raise ValueError(f"{self}.column_names has not been set")

        for curr_mo, next_mo in zip_longest(self.row_spec, self.row_spec[1:]):
            if curr_mo.column_count:
                pos_ = pos + curr_mo.column_count
            elif next_mo:
                try:
                    pos_ = self.column_names.index(next_mo.split, pos + 1)
                except ValueError as e:
                    raise ValueError(
                        f"split_on column {next_mo.split!r} for {next_mo} not found: "
                        f" (columns remaining are {self.column_names[pos + 1:]})"
                    ) from e
            else:
                pos_ = len(self.column_names)
            result.append(slice(pos, pos_))
            pos = pos_
        return result

    def _maprows(self, grouper, maker, rows) -> t.Iterator[t.Any]:
        if self.is_multi:
            object_rows = ([row[s] for s in self.split_points] for row in rows)
            if grouper:
                return grouper(object_rows)
            else:
                return (tuple(map(maker, enumerate(r))) for r in object_rows)
        else:
            return map(maker, zip(repeat(0), rows))


def group_by_and_join(
    mapped_column_names: list[tuple[str, ...]],
    row_spec: list[mapobject],
    join_spec: Sequence[joinedload],
    object_rows: t.Iterable[list[tuple]],
    key_columns: t.Optional[list[tuple[str]]] = None,
    _marker=object(),
):
    make_object = make_object_maker(mapped_column_names, row_spec)
    indexed_joins = translate_to_column_indexes(row_spec, join_spec)
    join_source_columns = {c.source_idx for c in indexed_joins}

    # Mapping of <column group index>: <currently loaded object>
    cur: dict[int, t.Any] = {}

    # List of column group indexes without backlinks: these are the top-level
    # objects we want to return
    output_columns = [n for n in range(len(row_spec)) if n not in join_source_columns]
    single_column = output_columns[0] if len(output_columns) == 1 else None
    items = None

    joined: set[tuple[int, int]] = set()
    emitted: set[tuple[int, int]] = set()

    output_row = {}
    output_queue = []

    for items, peek in pairwise(chain(object_rows, [None])):
        # Create objects from column data {{{
        for column_index, item in enumerate(items):
            if column_index in join_source_columns:
                if all(v is None for v in item):
                    cur[column_index] = make_object((column_index, NullObject))
                    continue
            ob = make_object((column_index, item))
            cur[column_index] = ob
        # }}}

        # Populate joins {{{
        for t_idx, attr, s_idx, cardinality, join_as_dict in indexed_joins:
            ob = cur[s_idx]
            dest = cur[t_idx]
            if dest is None:
                continue
            ob_key = (s_idx, id(dest), id(ob))
            if ob_key in joined:
                continue

            if cardinality == "*":
                if join_as_dict:
                    if attr in dest:
                        attrib = dest[attr]
                    else:
                        attrib = _marker
                else:
                    attrib = getattr(dest, attr, _marker)
                if attrib is _marker:
                    attrib = []
                    if join_as_dict:
                        dest[attr] = attrib
                    else:
                        setattr(dest, attr, attrib)
                if ob is not None:
                    attrib.append(ob)  # type: ignore
            else:
                if join_as_dict:
                    dest[attr] = ob
                else:
                    setattr(dest, attr, ob)
            joined.add(ob_key)
        # }}}
        if peek is None:
            changed = [(cur[ix], ix) for ix in output_columns]
        else:
            changed = [
                (cur[ix], ix)
                for ix in output_columns
                if (
                    items[ix] != peek[ix]
                    or all(i is None for i in items[ix])
                )
            ]

        if changed:
            if single_column is not None:
                for ob, col_index in changed:
                    if col_index == single_column:
                        yield ob
                        break
            else:
                for ob, col_index in changed:
                    ob_key = col_index, id(ob)
                    if ob_key in emitted:
                        continue
                    emitted.add(ob_key)
                    if col_index in output_row:
                        output_queue.append((ob, col_index))
                    else:
                        output_row[col_index] = ob
                if len(output_row) == len(output_columns):
                    yield tuple(output_row[ix] for ix in output_columns)
                    for ob2, pos2 in output_queue:
                        output_row[pos2] = ob2
                        yield tuple(output_row[ix] for ix in output_columns)
                    joined.clear()
                    emitted.clear()
                    output_row = {}
                    output_queue[:] = []


def translate_to_column_indexes(
    row_spec: list[mapobject], join_spec: Sequence[joinedload]
) -> list[column_indexes]:
    """
    :return:
        a list of tuples, one per join, representing:
        - the target object offset in the column list
        - the target attribute name
        - the source object offset in the column list
        - the join cardinality
        - a bool, if true the join should be dict-style (setitem) otherwise
          object style (setattr)
    """
    row_spec_indexes = {c.label: ix for ix, c in enumerate(row_spec)}

    def map_column(col: t.Any) -> int:
        if isinstance(col, int):
            return col
        return row_spec_indexes[col]

    result = []
    for j in join_spec:
        t_col = map_column(j.target)
        s_col = map_column(j.source)
        if t_col >= len(row_spec):
            raise ValueError(
                f"Target index {t_col} in join {j} exceeds number of mapped objects"
            )
        if s_col >= len(row_spec):
            raise ValueError(
                f"Source index {s_col} in join {j} exceeds number of mapped objects"
            )
        result.append(
            column_indexes(
                t_col, j.attr, s_col, j.cardinality, row_spec[t_col].join_as_dict
            )
        )
    return result


def make_object_maker(
    mapped_column_names: list[tuple[str, ...]],
    row_spec: list[t.Any],
) -> t.Callable[[t.Union[NullObjectType, tuple]], t.Any]:
    """
    Return a function that constructs the target type from a group of columns.

    The returned function will cache objects (the same input returns the
    same object) so that object identity may be relied on within the scope of a
    single query.
    """
    key_column_positions: list[list[int]] = []
    row_spec_cols = list(zip(row_spec, mapped_column_names))
    for mo, item_column_names in row_spec_cols:
        key_column_positions.append([])
        for c in mo.key_columns:
            try:
                key_column_positions[-1].append(item_column_names.index(c))
            except ValueError as e:
                import pprint

                mapped_columns_dump = {m.mapped: c for m, c in row_spec_cols}
                raise ValueError(
                    f"{c!r} specified in key_columns does not exist "
                    f"in the returned columns for {mo.mapped!r}. \n"
                    f"Mapped columns are: \n{pprint.pformat(mapped_columns_dump)}"
                ) from e

    def _object_maker() -> t.Generator[t.Any, tuple[int, tuple], None]:
        """
        Return a generator that accepts (rowspec_index, data) and yields
        the constructed object.
        """
        object_cache: dict[t.Any, t.Any] = {}
        ob = None

        # When loading multiple objects, ensure that proximate items loaded
        # with identical values map to the same object. This makes it possible
        # for joined loads to do the right thing, even if key_columns is not
        # set.
        row_cache: dict[t.Union[Sequence[t.Any], tuple[int, t.Any]], t.Any] = {}
        use_row_cache = len(row_spec) > 1
        mapping_items: list[tuple[t.Any, t.Any, bool, t.Optional[bool]]] = [
            (m.mapped, m.key_columns, m.positional, None) for m in row_spec
        ]

        i = -1
        rows_since_cache_flush = 0
        cache_hit_this_row = False
        cache_as_pickle_cols = {ix: False for ix in range(len(row_spec))}
        pickle = partial(dumps, protocol=HIGHEST_PROTOCOL)

        while True:
            (i, data) = yield ob
            cache_as_pickle = cache_as_pickle_cols[i]

            # Clear the row_cache when we find a full row with no cache hits.
            if i == 0 and row_cache and not cache_hit_this_row:
                cache_hit_this_row = False
                if rows_since_cache_flush < 2:
                    rows_since_cache_flush += 1
                else:
                    row_cache.clear()
                    rows_since_cache_flush = 0

            if data is NullObject:
                ob = None
                continue

            mapped, key_columns, positional, passthrough = mapping_items[i]
            if passthrough is None:
                passthrough = (
                    len(data) == 1
                    and isinstance(mapped, type)
                    and isinstance(data[0], mapped)
                )
                mapping_items[i] = (mapped, key_columns, positional, passthrough)
            if passthrough:
                ob = data[0]
            elif key_columns:
                key = (i, tuple(data[x] for x in key_column_positions[i]))
                if key in object_cache:
                    cache_hit_this_row = True
                    ob = object_cache[key]
                else:
                    ob = object_cache[key] = (
                        mapped(*data)
                        if positional
                        else mapped(**dict(zip(mapped_column_names[i], data)))
                    )
            elif use_row_cache:
                cache_key = pickle((i, data)) if cache_as_pickle else (i, data)
                try:
                    cache_hit = cache_key in row_cache
                except TypeError:
                    if cache_as_pickle:
                        raise
                    # data may contain unhashable types (eg postgresql ARRAY
                    # types), in which case a TypeError is thrown. Work around
                    # this by allowing keys for this column to be pickled.
                    cache_as_pickle_cols[i] = cache_as_pickle = True
                    cache_key = pickle(cache_key)
                    cache_hit = cache_key in row_cache

                if cache_hit:
                    cache_hit_this_row = True
                    ob = row_cache[cache_key]
                else:
                    ob = (
                        mapped(*data)
                        if positional
                        else mapped(**dict(zip(mapped_column_names[i], data)))
                    )
                    row_cache[cache_key] = ob
            else:
                ob = (
                    mapped(*data)
                    if positional
                    else mapped(**dict(zip(mapped_column_names[i], data)))
                )

    func = _object_maker()
    next(func)
    return func.send



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = [
    "Module",
    "Query",
    "joinone",
    "joinmany",
    "one_to_one",
    "one_to_many",
    "mapobject",
]

from .module import Module
from .query import Query
from .query import joinone
from .query import joinmany
from .query import one_to_one
from .query import one_to_many
from .mapobject import mapobject

__version__ = "5.2.1"



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class InvalidStatement(Exception):
    """
    The statement could not be parsed
    """


class NoResultFound(Exception):
    """
    A row was expected but the query did not generate one
    """


class MultipleResultsFound(Exception):
    """
    A single row was expected but the query returned multiple rows
    """


class ConnectionLimitError(Exception):
    """
    The connection pool has run out of available connections
    """


class Error(Exception):
    """
    The PEP-249 base error class
    """


class InterfaceError(Error):
    """
    PEP-249 error
    """


class DatabaseError(Error):
    """
    PEP-249 error
    """


class DataError(DatabaseError):
    """
    PEP-249 error
    """


class OperationalError(DatabaseError):
    """
    PEP-249 error
    """


class IntegrityError(DatabaseError):
    """
    PEP-249 error
    """


class InternalError(DatabaseError):
    """
    PEP-249 error
    """


class ProgrammingError(DatabaseError):
    """
    PEP-249 error
    """


class NotSupportedError(DatabaseError):
    """
    PEP-249 error
    """


pep_249_exceptions = [
    Error,
    InterfaceError,
    DatabaseError,
    DataError,
    OperationalError,
    IntegrityError,
    InterfaceError,
    ProgrammingError,
    NotSupportedError,
]
pep_249_exception_names = {c.__name__: c for c in pep_249_exceptions}



from collections.abc import Sequence, Mapping
import typing as t

Parameters = t.Union[Sequence[t.Any], Mapping[str, t.Any]]
Row = Sequence[t.Any]

CursorDescription = Sequence[
    #: name, type_code, display_size, internal_size, precision, scale, null_ok
    tuple[
        str,
        object,
        t.Optional[int],
        t.Optional[int],
        t.Optional[int],
        t.Optional[int],
        t.Optional[bool],
    ]
]


class Connection(t.Protocol):
    #: Optional extension
    messages: list[tuple[type, Exception]]

    #: Optional extension
    autocommit: bool

    #: Optional extension
    errorhandler: t.Optional[
        t.Callable[["Connection", "Cursor", t.Type[Exception], Exception], t.NoReturn]
    ]

    def close(self) -> None:
        ...

    def commit(self) -> None:
        ...

    def rollback(self) -> None:
        ...

    def cursor(self) -> "Cursor":
        ...


class Cursor(t.Protocol):
    description: t.Optional[CursorDescription]
    rowcount: int
    arraysize: int

    #: Optional extension
    rownumber: t.Optional[int]

    #: Optional extension
    connection: Connection

    #: Optional extension
    messages: list[tuple[type, Exception]]

    #: Optional extension
    lastrowid: t.Any

    #: Optional extension
    errorhandler: t.Optional[
        t.Callable[["Connection", "Cursor", t.Type[Exception], Exception], t.NoReturn]
    ]

    def callproc(
        self, procname: str, parameters: t.Optional[Sequence[t.Any]] = None
    ) -> None:
        "Optional extension"

    def close(self) -> None:
        ...

    def execute(
        self, operation: str, parameters: Parameters = None
    ) -> None:
        ...

    def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters]
    ) -> None:
        ...

    def fetchone(self) -> Row | None:
        ...

    def fetchmany(self, size: int = 1) -> Sequence[Row]:
        ...

    def fetchall(self) -> Sequence[Row]:
        ...

    def nextset(self) -> None:
        "Optional extension"

    def setinputsizes(self, sizes) -> None:
        ...

    def setoutputsize(self, size: int, column: t.Optional[int]) -> None:
        ...

    def scroll(self, value: int, mode: str = "relative") -> None:
        "Optional extension"

    def next(self) -> Row:
        "Optional extension"
        ...

    def __iter__(self) -> t.Iterator[Sequence[t.Any]]:
        "Optional extension"
        ...



import re
import collections.abc
import inspect
import typing as t
from itertools import count
import pickle

from .module import Module
from .parsing import parse_type_hints


def template(base_imports, project_imports, queries):
    return render_template(
        """
        # Auto-generated by embrace.codegen
        {% for imp, alias in sorted(bi.items()) %}
        import {{ imp }} as {{ alias }}
        {% end %}

        {% if pi %}
        if __import__("typing").TYPE_CHECKING:
            {% for imp, alias in sorted(pi.items()) %}
            import {{ imp }} as {{ alias }}
            {% end %}
        {% end %}

        _m = {{ bi.embrace_module }}.Module()

        {% for q in queries %}class _{{ q.name }}_Query({{ bi.embrace_query }}.Query[{{q.return_annotation}}, {{ q.scalar_return_annotation }}]):
            def resultset(
                {% capture args %}self, conn: {{ bi.typing }}.Optional[{{ bi.embrace_types }}.Connection], {% for n, h in q.args %}{{ n }}: {{h}}, {% end %}*, params: {{ bi.typing }}.Optional[{{bi.collections_abc}}.Mapping]=None, debug: bool=False, _module = _m{% endcapture %}
            ) -> _Q.ResultSet[{{ q.return_annotation }}, {{ q.scalar_return_annotation }}]:
                return super().resultset(
                    {% capture callargs %}conn, {% for n in q.arg_names %}{{ n }}={{ n }}, {% end %} params=params, debug=debug {% endcapture %}
                )
            def one({{ args }}) -> {{ q.return_annotation }}:
                return super().one({{ callargs }})
            def first({{ args }}) ->  _T.Union[{{ q.return_annotation }}, None]:
                return super().first({{ callargs }})
            def many({{ args }}) -> _C.Iterator[{{ q.return_annotation }}]:
                return super().many({{ callargs }})
            def column({{ args }}) -> _C.Iterator[{{ q.scalar_return_annotation }}]:
                return super().column({{ callargs }})
            def scalar({{ args }}, **_kw) -> {{ q.scalar_return_annotation }}:
                return super().scalar({{ callargs }}, **_kw)

        {% end %}

        {% for q in queries %}
        _m.queries["{{q.name}}"] = {{ q.name }} = _{{ q.name }}_Query()
        {{ q.name }}.__dict__ = {{ bi.pickle }}.loads({{ repr(q.pickle) }})
        {% end %}

        def __getattr__(name):
            return getattr(_m, name)
        """,
        {"bi": base_imports, "pi": project_imports, "i": base_imports | project_imports, "queries": queries},
    )


def generate_module_source(module):
    queries = []

    # Imports required by embrace's internals
    base_imports = ImportDict({
        "pickle": "_P",
        "typing": "_T",
        "collections.abc": "_C",
        "embrace.types": "_ET",
        "embrace.module": "_M",
        "embrace.query": "_Q",
    })
    # Imports imposed by the project, these will be wrapped in an
    # ``if TYPE_CHECKING`` statement to remove the possibility of circular
    # imports
    project_imports = ImportDict({})

    symbol_aliases = symbol_aliaser()
    for q in sorted(module, key=lambda q: q.name or ""):
        q.resolve_includes(module)
        md = q.metadata
        placeholder_info = q.get_placeholder_info()
        if placeholder_info:
            arg_names, placeholder_hints = zip(*placeholder_info)
            arg_hints = symbol_aliases(
                base_imports, project_imports, parse_type_hints(md.imports, placeholder_hints)
            )
        else:
            arg_names = arg_hints = []

        if q.row_mapper:
            row_spec = q.row_mapper.row_spec_after_joins()
        else:
            row_spec = []

        if row_spec:
            mapped = symbol_aliases(base_imports, project_imports, (mo.mapped for mo in row_spec))
            if len(mapped) > 1:
                return_annotation = f"""tuple[{', '.join(f'"{m}"' for m in mapped)}]"""
            else:
                return_annotation = f'"{mapped[0]}"'
            scalar_return_annotation = f'"{mapped[0]}"'
        else:
            return_annotation = "tuple"
            scalar_return_annotation = f"{base_imports.typing}.Any"

        queries.append(
            AttrDict(
                q.__dict__
                | {
                    "args": list(zip(arg_names, arg_hints)),
                    "arg_names": arg_names,
                    "return_annotation": return_annotation,
                    "scalar_return_annotation": scalar_return_annotation,
                    "pickle": pickle.dumps(q.__dict__),
                }
            )
        )

    return template(base_imports, project_imports, queries)


def symbol_aliaser() -> (
    t.Callable[
        [
            collections.abc.Mapping[str, t.Optional[str]],
            collections.abc.MutableMapping[str, t.Optional[str]],
            t.Iterator
        ], list[str]
    ]
):
    c = count()

    def symbol_aliases(
        base_aliases: collections.abc.Mapping[str, t.Optional[str]],
        project_aliases: collections.abc.Mapping[str, t.Optional[str]],
        obs: t.Iterator[t.Any],
    ) -> list[str]:

        symbols = []
        for ix, ob in enumerate(obs):
            if ob is None:
                symbols.append(f"{base_aliases['typing']}.Any")
                continue
            try:
                ra = t.get_origin(inspect.signature(ob).return_annotation)
                if ra is not inspect._empty and ra is not None:
                    ob = ra
            except (TypeError, ValueError):
                pass
            try:
                mod = ob.__module__
            except AttributeError:
                symbols.append(f"{base_aliases['typing']}.Any")
            else:
                name = getattr(ob, "__qualname__", None)
                if name is None or "<locals>" in name or mod == "embrace.mapobject":
                    # <locals>: non-exportable symbol,
                    # typically the result of calling mapobject.dataclass().
                    #
                    # embrace.mapobject:
                    # special case this because we can't statically infer the type
                    # of mapobject.passthrough
                    #
                    # In both cases typing.Any is the best we can do.
                    symbols.append(f"{base_aliases['typing']}.Any")
                elif mod == "builtins":
                    symbols.append(name)
                else:
                    if mod not in project_aliases:
                        project_aliases[mod] = f"_{next(c)}"
                    symbols.append(f"{project_aliases[mod]}.{name}")
        return symbols

    return symbol_aliases


class AttrDict(dict):
    def __getattr__(self, name):
        return self[name]


class ImportDict(dict):
    def __getattr__(self, name):
        return super().__getitem__(name.replace("_", "."))


def render_template(
    template: str,
    context: collections.abc.Mapping,
):
    tag_pattern = re.compile(
        r"""
        (
            {{ .*? }}
            | {% (?:[^%]|%[^}])* %}(?:[ \t]*\n)
            | {% (?:[^%]|%[^}])* %}
        )
        """,
        re.DOTALL | re.VERBOSE
    )
    template = dedent(template)
    tokens = [token for token in tag_pattern.split(template) if token]
    code = _parse_tokens(tokens)
    ns = dict(context)
    exec(code, ns)
    return "".join(ns["__output__"][0])


def _parse_tokens(
    tokens: collections.abc.Sequence[str],
    substitution_pattern=re.compile(r"{{\s*(.*?)\s*}}"),
    block_pattern=re.compile(
        r"{%\s(if |for |elif |else|capture |endcapture|end)(.*?)\s*%}\s*"
    ),
) -> str:
    captures = []
    lines = [
        "__output__ = [[]]\n"
        "def __out(s):\n"
        "    __output__[-1].append(str(s))\n"
        "\n",
    ]
    indent = 0

    def add_line(line, indent_delta=0, newline=True):
        nonlocal indent
        if lines[-1].endswith(":\n"):
            indent += 1
        indent += indent_delta
        if indent < 0:
            raise SyntaxError("Unexpected end tag")
        lines.append(f"{'    ' * indent}{line}\n")

    for token in tokens:
        indent_delta = 0
        if mo := substitution_pattern.match(token):
            add_line(f"__out({mo.group(1)})")

        elif mo := block_pattern.match(token):
            command, content = mo.groups()
            if command in {"elif ", "else", "end"}:
                indent_delta = -1
            if command == "end":
                indent -= 1
            elif command == "capture ":
                captures.append(content)
                add_line("__output__.append([])")
            elif command == "endcapture":
                capture = captures.pop()
                add_line(f"{capture} = ''.join(__output__.pop())")
                add_line(f"__out({capture})")
            else:
                add_line(f"{command}{content}:", indent_delta)
        else:
            add_line(f"__out({str(token)!r})")

    return "".join(lines)


def dedent(s: str):
    indent = None
    result = []
    for line in s.split("\n"):
        if indent is None:
            if line.strip():
                indent = re.match(r"^\s*", line).group(0)  # type: ignore
        if indent is None:
            result.append(line)
        else:
            result.append(line.removeprefix(indent))
    return "\n".join(result)


if __name__ == "__main__":
    import os
    import pathlib
    import argparse
    import tempfile

    argparser = argparse.ArgumentParser(
        description="Compile SQL queries into a single python module"
    )
    argparser.add_argument("--output", "-o", default="-", type=pathlib.Path)
    argparser.add_argument("dir", nargs="+")
    args = argparser.parse_args()
    m = Module()
    for d in args.dir:
        m.load_dir(d)
    s = generate_module_source(m)
    if args.output != "-":
        tmp = tempfile.NamedTemporaryFile(
            "w", encoding="utf-8", dir=args.output.parent, delete=False
        )
        tmp.write(s)
        tmp.close()
        os.rename(tmp.name, args.output)
    else:
        print(s)



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import dataclasses
from itertools import count
from functools import partial
from functools import cache
from typing import Any
from typing import Iterator
from typing import Callable
from typing import Mapping
from typing import Union
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING
import typing as t

if TYPE_CHECKING:
    from embrace.query import Query

import sqlparse
import sqlparse.sql

from .mapobject import mapobject
from .mapobject import make_rowspec
from .mapobject import joinedload

BindParams = Union[tuple, Mapping]
Tokens = list[tuple]

param_styles = {
    "qmark": (True, lambda name, c: "?"),
    "numeric": (True, lambda name, c: f":{c}"),
    "format": (True, lambda name, c: "%s"),
    "pyformat": (False, lambda name, c: f"%({name})s"),
    "named": (False, lambda name, c: f":{name}"),
}
name_pattern = re.compile(r":name\s+(?P<name>\w+)", re.X)
args_pattern = re.compile(r":args\s+(?P<args>.*)$", re.X)
return_pattern = re.compile(r":returns?\s+(?P<returns>.*)$", re.X)
joins_pattern = re.compile(r":join(?P<joins>(?:one|many)\s+.*)$", re.X)
split_on_pattern = re.compile(r":spliton\s+(?P<split_on>.*)$", re.X)
key_columns_pattern = re.compile(r":keycolumns?\s+(?P<key_columns>.*)$", re.X)
import_pattern = re.compile(r":(?P<imports>(?:import|from)\s+.*)$", re.X)
include_pattern = re.compile(r":include:([a-zA-Z_]\w*)")
param_pattern = re.compile(
    r"""
    # Don't match if:
    # - preceded by backslash (an escape) or ':' (an SQL cast, eg '::INT')
    # - it's an include pattern
    (?<![:\\])
    (?!:include:)
    (?<!:include:)

    # Optional parameter type
    (?:
        :(?P<param_type>
            value|v|value\*|v\*|tuple|t|tuple\*|t\*|identifier|i|raw|r
        )
    )?

    # An identifier
    :(?P<param>[a-zA-Z_]\w*)

    # Optional type hint
    (?::(?P<type_hint>[a-zA-Z\[\]_.|]+))?

    # followed by a non-word char, or end of string
    (?=\W|$)
    """,
    re.X,
)


class ParseError(Exception):
    ...


@dataclasses.dataclass
class Metadata:
    name: Optional[str] = None
    lineno: Optional[int] = None
    args: Optional[str] = None
    returns: Optional[str] = None
    imports: list[str] = dataclasses.field(default_factory=list)
    joins: list[str] = dataclasses.field(default_factory=list)
    split_on: list[str] = dataclasses.field(default_factory=list)
    key_columns: list[str] = dataclasses.field(default_factory=list)

    def update(self, other: "Metadata"):
        defaults = self.__class__().asdict()
        for k, v in other.asdict().items():
            if isinstance(getattr(self, k), list):
                getattr(self, k).extend(v)
            else:
                if v != defaults[k]:
                    setattr(self, k, v)

    def asdict(self):
        return dataclasses.asdict(self)


def parse_comment_metadata(s: str) -> Metadata:
    lines = (re.sub(r"^\s*--", "", line).strip() for line in s.split("\n"))
    patterns = [
        name_pattern,
        args_pattern,
        return_pattern,
        import_pattern,
        joins_pattern,
        split_on_pattern,
        key_columns_pattern,
    ]
    metadata = Metadata()
    for line in lines:
        for p in patterns:
            if mo := p.match(line):
                for k, v in mo.groupdict().items():
                    if v is not None:
                        if isinstance(getattr(metadata, k), list):
                            getattr(metadata, k).append(v)
                        else:
                            setattr(metadata, k, v)

    return metadata


def quote_ident_ansi(s):
    s = str(s)
    if "\x00" in s:
        raise ValueError(
            "Quoted identifiers can contain any character, "
            "except the character with code zero"
        )
    return f'''"{s.replace('"', '""')}"'''


def compile_includes(
    statement: str,
) -> tuple[list[str], Callable[[Sequence["Query"]], str]]:
    """
    Parse a statement str, looking for ':include:<name>' patterns. Return the
    list of include names found, and a function that will accept the named
    queries in order and return a modified statement string.
    """

    # Will be [<stmt>, <include_name>, <stmt>, <include_name>, … <stmt>]
    splits = include_pattern.split(statement)

    # List of matched include names
    names = splits[1::2]

    def strip_semicolon(stmt: str) -> str:
        return stmt.rstrip().rstrip(";")

    def replace_includes(includes: Sequence["Query"]) -> str:
        iterincludes = iter(includes)
        return "".join(
            s
            if ix % 2 == 0
            else ";".join(map(strip_semicolon, next(iterincludes).statements))
            for ix, s in enumerate(splits)
        )

    return names, replace_includes


def compile_bind_parameters(
    target_style: str, sql: str, bind_parameters: Mapping
) -> tuple[str, BindParams]:
    """
    :param target_style: A DBAPI paramstyle value (eg 'qmark', 'format', etc)
    :param sql: An SQL str
    :bind_parameters: A dict of bind parameters for the query

    :return: tuple of `(sql, bind_parameters)`. ``sql`` will be rewritten with
             the target paramstyle; ``bind_parameters`` will be a tuple or
             dict as required.
    :raises: TypeError, if the bind_parameters do not match the query
    """

    is_positional, param_gen = param_styles[target_style]

    if target_style[-6:] == "format":
        sql = sql.replace("%", "%%")
    params: Union[list, dict[str, Any]]
    if is_positional:
        params = []
    else:
        params = {}

    transformed_sql = param_pattern.sub(
        partial(
            replace_placeholder,
            params,
            param_gen,
            count(1),
            bind_parameters,
        ),
        sql,
    )
    if is_positional:
        return transformed_sql, tuple(params)
    return transformed_sql, params  # type: ignore


def get_placeholder_info(sql: str) -> Iterator[tuple[str, str]]:
    """
    Generate tuples of ``(<name>, <optional type hint>)``
    for all placeholders present in ``sql``
    """
    for match in param_pattern.finditer(sql):
        gd = match.groupdict()
        yield gd["param"], gd["type_hint"] or "None"


@cache
def split_statements(
    sql: str,
) -> list[tuple[Metadata, list[sqlparse.sql.Statement]]]:
    """
    Split an sql string into multiple queries and parse any metadata
    found in comments
    """
    statements = sqlparse.parse(sql)
    result = []
    positions = []
    offset = 0
    for s in statements:
        offset = sql.index(str(s), offset)
        positions.append(offset)

    statements = ((pos, s) for pos, s in zip(positions, statements) if str(s).strip())
    current: tuple[Optional[Metadata], list[sqlparse.sql.Statement]] = (None, [])

    for pos, statement in statements:
        metadata, tokens = parse_tokens(statement)
        metadata.lineno = sql[:pos].count("\n") + 1
        statement.tokens[:] = tokens

        if starts_new_statement(metadata):
            if current[0] is not None:
                result.append(current)
            current = metadata, [statement]

        else:
            if current[0] is None:
                current = (metadata, [statement])
            else:
                current[0].update(metadata)
                current[1].append(statement)

    if current[0] is not None:
        result.append(current)
    return result


def replace_placeholder(
    params: Union[dict[str, Any], list[Any]],
    param_gen,
    placeholder_counter,
    bind_parameters,
    match,
):
    group = match.groupdict().get
    pt = group("param_type")
    p = group("param")
    if pt in {None, "value", "v"}:
        if isinstance(params, list):
            params.append(bind_parameters[p])
        else:
            params[p] = bind_parameters[p]
        return param_gen(p, next(placeholder_counter))  # type: ignore
    elif pt in {"raw", "r"}:
        return str(bind_parameters[p])
    elif pt in {"identifier", "i"}:
        return quote_ident_ansi(bind_parameters[p])
    elif pt in {"value*", "v*", "tuple", "t"}:
        # :tuple: -> (?, ?, ?)
        # :value*: -> ?, ?, ?

        placeholder = make_sequence_placeholder(
            params, param_gen, placeholder_counter, list(bind_parameters[p])
        )
        if pt in {"tuple", "t"}:
            return f"({placeholder})"
        return placeholder
    elif pt in {"tuple*", "t*"}:
        # :tuple*: -> (?, ?, ?), (?, ?, ?), …
        return ", ".join(
            f"({make_sequence_placeholder(params, param_gen, placeholder_counter, list(items))})"
            for items in bind_parameters[p]
        )
    else:
        raise ValueError(f"Unsupported param_type {pt}")


def make_sequence_placeholder(
    params: Union[list[Any], dict[str, Any]],
    param_gen: Callable[[str, int], str],
    placeholder_counter: Iterator[int],
    items: list,
) -> str:
    """
    Return a placeholder for a sequence of parameters, in the format::

        ?, ?, ?, ...

    Or::

        :_1, :_2, :_3, ...

    Modify ``params`` in place to include the placeholders.
    """
    numbers = [next(placeholder_counter) for ix in range(len(items))]
    names = [f"_{n}" for n in numbers]
    if isinstance(params, list):
        params.extend(items)
    else:
        params.update(zip(names, items))
    return ", ".join(param_gen(name, c) for name, c in zip(names, numbers))


def starts_new_statement(metadata: Metadata) -> bool:
    return bool(metadata.name)


def parse_tokens(statement: sqlparse.sql.Statement) -> tuple[Metadata, Tokens]:
    """
    Parse out any metadata from an sqlparse statement
    Return a Metadata object and a new list of tokens with metadata removed
    """
    new_tokens = []
    metadata: Metadata = Metadata()
    for token in statement.tokens:
        if isinstance(token, sqlparse.sql.Comment) or (
            token.ttype and token.ttype[0] == "Comment"
        ):
            metadata.update(parse_comment_metadata(str(token)))
        else:
            new_tokens.append(token)
    return metadata, new_tokens


def parse_type_hints(imports: Sequence[str], type_hints: Sequence[str]) -> list[t.Any]:
    imports_str = "\n".join(imports)
    code = (
        f"{imports_str}\n"
        f"type_hints = [{', '.join(type_hints)}]"
    )
    _code = compile(code, "<string>", "exec")
    ns = {}
    try:
        exec(_code, ns)
    except NameError as e:
        raise NameError(
            f"{e.args[0]} (do you need to add `-- :import {e.name}`?)", name=e.name
        ) from e
    return ns["type_hints"]


@dataclasses.dataclass
class ParsedReturning:
    row_spec: list[mapobject] = dataclasses.field(default_factory=list)
    imports: list[str] = dataclasses.field(default_factory=list)
    joins: list[joinedload] = dataclasses.field(default_factory=list)
    key_columns: list[tuple[str, ...]] = dataclasses.field(default_factory=list)
    split_on: list[str] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return bool(self.row_spec)


def parse_returning_metadata(
    imports: list[str] = [],
    returns: t.Optional[str] = None,
    joins: list[str] = [],
    key_columns: list[str] = [],
    split_on: list[str] = [],
) -> ParsedReturning:
    """
    Parse the query returning metadata into a ParsedReturning object,
    resolving imported symbols into live objects
    """
    result = ParsedReturning()
    if imports or returns:
        imports.append("from embrace import mapobject")
        imports_str = "\n".join(imports)
        code = (
            f"{imports_str}\n"
            f'returns = {returns if returns and returns.strip() else "None"}'
        )
        ns = {}
        _code = compile(code, "<string>", "exec")
        try:
            exec(_code, ns)
        except NameError as e:
            import textwrap
            code = textwrap.indent(code, "  ")
            raise NameError(
                f"{e.args[0]} "
                f"(do you need to add `-- :import {e.name}`?)\n"
                f"Generated code is:\n\n{code}",
                name=e.name
            ) from e
    else:
        ns = {"returns": None}
    rs = ns["returns"]
    rs_origin = t.get_origin(rs)
    if rs is None:
        result.row_spec = []
    elif rs_origin is tuple:
        rs_args = t.get_args(rs)
        if rs_args == (Ellipsis,):
            result.row_spec = make_rowspec(
                tuple, split_on=[], key_columns=[], positional=False
            )
        elif Ellipsis in rs_args:
            raise ParseError(
                f"Invalid return annotation {returns}: tuples must be completely specified."
            )
        else:
            result.row_spec = make_rowspec(
                t.get_args(rs), split_on=[], key_columns=[], positional=False
            )
    elif rs_origin is None:
        result.row_spec = make_rowspec(
            rs, split_on=[], key_columns=[], positional=False
        )
    else:
        raise ParseError(
            f"Invalid return annotation {returns}: unsupported typing origin {rs_origin}"
        )

    result.joins = parse_joins(ns, joins)
    result.key_columns = [tuple(s.strip() for s in k.split(",")) for k in key_columns]
    result.split_on = split_on

    return result


def parse_joins(
    ns: dict[str, Any],
    joins: list[str],
    cardinality_pattern=re.compile(r"^(one|many)\s*(.*)$"),
) -> list[joinedload]:
    result: list[joinedload] = []
    for s in joins:
        if mo := cardinality_pattern.match(s):
            cardinality = mo.group(1)
            s = mo.group(2).strip()
        else:
            cardinality = "one"
            s = s.strip()
        lhs, rhs = re.split(r"\s*=\s*", s)
        try:
            lhs, lhs_attr = lhs.rsplit(".", 1)
        except ValueError:
            raise ParseError(
                f"{lhs!r} should be a dotted path in join expression {s!r}"
            )
        lhs_ob = load_join_symbol(ns, s, lhs)
        rhs_ob = load_join_symbol(ns, s, rhs)
        result.append(
            joinedload(
                lhs_ob,
                lhs_attr,
                rhs_ob,
                "1" if cardinality == "one" else "*",
            )
        )

    return result


def load_join_symbol(ns: dict[str, t.Any], joinexpr: str, dottedpath: str) -> t.Any:
    """
    Load a symbol from dict `ns` addressed by a dotted path
    """
    names = dottedpath.split(".")
    ob = None
    for name in names:
        if ob is None:
            next_ob = ns.get(name, name)
            if next_ob is name:
                if len(names) == 1:
                    return name
                raise ParseError(
                    f"name {name!r} is not defined in join expression {joinexpr!r}. "
                    f"Do you need to add ':import {name}'?"
                )
        else:
            try:
                next_ob = getattr(ob, name)
            except AttributeError:
                raise ParseError(
                    f"{type(ob)} object has no attribute {name!r} "
                    f"in join expression {joinexpr!r}"
                )
        ob = next_ob
    if ob is None:
        raise ParseError(f"Bad join expression {joinexpr!r}")
    return ob



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pathlib
import dataclasses
from itertools import count
from typing import overload
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import Tuple
from typing import Union

import wrapt

from .query import Query
from .types import Connection
from .exceptions import InvalidStatement
from .parsing import split_statements

SQL_FILE_GLOB = "**/*.sql"


class Module:
    _conn = None
    queries: dict = {}
    _query_mtimes: dict[pathlib.Path, float] = {}
    directories: list[pathlib.Path]

    def __init__(self, *paths: Union[str, pathlib.Path], auto_reload=False):
        self.queries = {}
        self.directories = []
        self.auto_reload = auto_reload
        for p in paths:
            self.load_dir(p)

    def __getattr__(self, name):
        if self.auto_reload:
            self.ensure_up_to_date(name)
            try:
                query = reloadable_query_proxy(self, name)
            except KeyError:
                raise AttributeError(name)
        else:
            try:
                query = self.queries[name]
            except KeyError:
                import difflib

                matches = difflib.get_close_matches(name, list(self.queries), n=1)
                raise AttributeError(
                    f"{self!r} has no attribute {name!r}"
                    + (f" Did you mean {matches[0]!r}?" if matches else "")
                )
        if not query.includes_resolved:
            query.resolve_includes(self)
        return query

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return

    def __iter__(self):
        return iter(self.queries.values())

    def __repr__(self):
        return f"<{self.__class__.__module__}.{self.__class__.__qualname__} {':'.join(map(str, self.directories))}>"

    def clear(self):
        self.queries.clear()

    def add_query(self, name, query: Union[Query, str]):
        if isinstance(query, str):
            return self.add_query(name, Query(query))
        if name in self.queries:
            raise InvalidStatement(
                f"Can't add query {name!r} in {query.source}: "
                f"{self!r} already has an attribute named {name!r} "
                f"(loaded from {getattr(self, name).source})"
            )

        if self._conn:
            query = query.bind(self._conn)
        query.name = name
        self.queries[name] = query
        if query.source:
            path = pathlib.Path(query.source)
            if path.exists():
                self._query_mtimes[path] = path.stat().st_mtime

    def add_queries(self, qs: Union[Iterable[Tuple[str, Query]], Mapping[str, Query]]):
        if isinstance(qs, Mapping):
            qs = qs.items()
        for name, query in qs:
            self.add_query(name, query)

    def load_dir(self, path: Union[str, pathlib.Path]):
        path = pathlib.Path(path)
        if path not in self.directories:
            self.directories.append(path)
        for p in path.glob(SQL_FILE_GLOB):
            self.load_file(p)

        return self

    def load_file(self, path: pathlib.Path):
        queries = list(load_queries(path))
        for query in queries:
            self.add_query(query.name, query)

    def ensure_up_to_date(self, query_name: str) -> bool:
        """
        Ensure that the named query is up to date, reloading from disk if
        required.

        Returns ``True`` if the query was reloaded
        """
        try:
            query = self.queries[query_name]
        except KeyError:
            self.clear()
            for path in self.directories:
                self.load_dir(path)
            return True
        includes_reloaded = False
        for included in query.includes:
            if self.ensure_up_to_date(included):
                includes_reloaded = True
        path = pathlib.Path(query.source)
        try:
            last_mtime = self._query_mtimes[path]
        except KeyError:
            last_mtime = 0.0

        if not path.exists():
            mtime = 0.0
        else:
            mtime = path.stat().st_mtime

        if mtime != last_mtime:
            for item in list(self.queries):
                if self.queries[item].source == query.source:
                    del self.queries[item]

            if path.exists():
                self.load_file(path)
                self._query_mtimes[path] = mtime
            return True
        if includes_reloaded:
            query.resolve_includes(self)
            return True
        return False

    def query(self, sql: str) -> Query:
        """
        Return a new, unnamed query.
        This convenience method is intended to facilitate ad-hoc
        queries, for example::

            >>> mod = embrace.module.Module()
            >>> result = mod.query("SELECT * FROM mytable").one(conn)

        It differs from instantiating ``embrace.query.Query``
        directly in that:
        - ``:include:`` markers will be resolved,
        - if the module has a connection bound,
          it will be bound to the returned query
        """
        query = Query(sql, source="<string>")
        if self._conn:
            query = query.bind(self._conn)
        query.resolve_includes(self)
        return query

    @overload
    def _run_query(self, result_type: str, sql: str, *args, **kwargs):
        ...

    @overload
    def _run_query(self, result_type: str, conn: Connection, sql: str, *args, **kwargs):
        ...

    def _run_query(self, result_type, *args, **kwargs):
        if isinstance(args[0], str):
            conn = self._conn
            sql, *args = args
        else:
            conn, sql, *args = args
        query = self.query(sql)
        return getattr(query, result_type)(conn, *args, **kwargs)

    def resultset(self, *args, **kwargs):
        return self._run_query("resultset", *args, **kwargs)

    def execute(self, *args, **kwargs):
        return self._run_query("execute", *args, **kwargs)

    def one(self, *args, **kwargs):
        return self._run_query("one", *args, **kwargs)

    def one_or_none(self, *args, **kwargs):
        return self._run_query("one_or_none", *args, **kwargs)

    def first(self, *args, **kwargs):
        return self._run_query("first", *args, **kwargs)

    def many(self, *args, **kwargs):
        return self._run_query("many", *args, **kwargs)

    def scalar(self, *args, **kwargs):
        return self._run_query("scalar", *args, **kwargs)

    def affected(self, *args, **kwargs):
        return self._run_query("affected", *args, **kwargs)

    def column(self, *args, **kwargs):
        return self._run_query("column", *args, **kwargs)

    def cursor(self, *args, **kwargs):
        return self._run_query("cursor", *args, **kwargs)

    def bind(self, conn) -> "Module":
        """
        Return a copy of the module bound to a database connection
        """
        cls = self.__class__
        bound = cls.__new__(cls)
        bound.__dict__ = {
            "_conn": conn,
            "auto_reload": self.auto_reload,
            "directories": self.directories,
            "_query_mtimes": self._query_mtimes,
            "queries": {name: q.bind(conn) for name, q in self.queries.items()},
        }
        return bound

    def transaction(self, conn) -> "Transaction":
        return Transaction(self, conn)

    def savepoint(self, conn) -> "Savepoint":
        return Savepoint(self, conn)


class Transaction:
    def __init__(self, module, conn):
        self.conn = conn
        self.module = module.bind(conn)

    def __enter__(self):
        return self.module

    def __exit__(self, type, value, traceback):
        if type:
            self.conn.rollback()
        else:
            self.conn.commit()

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()


class Savepoint:
    _seq = count()

    def __init__(self, module, conn):
        self.conn = conn
        self.module = module.bind(conn)
        self.savepoint = f"sp_{next(self._seq)}"
        self._cursor = None

    def __enter__(self):
        self._cursor = self.conn.cursor()
        self._cursor.execute(f"SAVEPOINT {self.savepoint}")
        return self.module

    def __exit__(self, type, value, traceback):
        if self._cursor is not None:
            if type:
                self._cursor.execute(f"ROLLBACK TO SAVEPOINT {self.savepoint}")
            else:
                self._cursor.execute(f"RELEASE SAVEPOINT {self.savepoint}")
            self._cursor.close()


def load_queries(path: pathlib.Path) -> Iterator[Query]:
    with path.open("r", encoding="UTF-8") as f:
        sql = f.read()
        for ix, (metadata, statements) in enumerate(split_statements(sql)):
            if not metadata.name:
                if ix == 0:
                    metadata = dataclasses.replace(metadata, name=path.stem)
                else:
                    raise InvalidStatement(
                        f"{path!s}, line {metadata.lineno}: no name specified (eg `-- :name my_query_name`)"
                    )
            yield Query.from_statements(
                metadata,
                statements,
                source=str(path),
                lineno=metadata.lineno,
                name=metadata.name,
            )


def reloadable_query_proxy(module, name):
    def get_query():
        return module.queries[name]

    class ReloadableQueryProxy(wrapt.ObjectProxy):
        def __call__(self, *args, **kwargs):
            module.ensure_up_to_date(name)
            self.__wrapped__ = get_query()
            return self.__wrapped__(*args, **kwargs)

        def bind(self, conn):
            return reloadable_query_proxy(module, name)

    return ReloadableQueryProxy(get_query())
